intersect_tri: ; Moeller-Trumbore method
; in:
;     xmm0 - ray direction  ; should be normalized
;     xmm1 - ray orgin
;     xmm2 - tri vert1
;     xmm3 - tri vert2
;     xmm4 - tri vert3
;     if  eax = 1 - intersction with edge
;        xmm6 - edge lenght
;     if  eax = 0 - intersect with ray (classic)
; out:
;     eax  = 1 - intersection occured
;     xmm0 - float lo -> hi = t, v, u, ...
      push    ebp
      mov     ebp,esp
      and     ebp,-16
      sub     esp,220

       .dir    equ [ebp-16]
       .origin equ [ebp-32]
       .ta     equ [ebp-48]
       .tb     equ [ebp-64]
       .tc     equ [ebp-80]
       .tvec   equ [ebp-96]
       .pvec   equ [ebp-112]
       .qvec   equ [ebp-128]
       .e1     equ [ebp-128-16]
       .ift    equ dword[ebp-152]
       .invdet equ [ebp-156]
       .det    equ [ebp-160]
       .ed_l   equ [ebp-164]
       .u      equ [ebp-168]
       .v      equ [ebp-172]
       .t      equ [ebp-176]
       .e2     equ [ebp-192]

      movaps   .dir,xmm0
      movaps   .origin,xmm1
      movaps   .ta,xmm2
      movaps   .tb,xmm3
      movaps   .tc,xmm4
      mov      .ift,eax
      movss    .ed_l,xmm6
      subps    xmm3,xmm2
      subps    xmm4,xmm2
      andps    xmm3,[zero_hgst_dd]
      andps    xmm4,[zero_hgst_dd]
      movaps   .e1,xmm3
      movaps   .e2,xmm4

      lea      esi,.dir
      lea      edi,.e2
      lea      ebx,.pvec
      call     cross_aligned

      movaps   xmm0,.e1
      mulps    xmm0,.pvec
 ;     andps    xmm0,[zero_hgst_dd]
      haddps   xmm0,xmm0
      haddps   xmm0,xmm0
      movss    .det,xmm0
;      cmpnless xmm0,[eps]
;      movd     eax,xmm0
;      or       eax,eax
;      jz       @f
      comiss   xmm0,[eps]
      jl       @f

      rcpss    xmm0,.det
      movss    .invdet,xmm0

      movaps   xmm0,.origin
      subps    xmm0,.ta
      andps    xmm0,[zero_hgst_dd]
      movaps   .tvec,xmm0

      mulps    xmm0,.pvec
      haddps   xmm0,xmm0
      haddps   xmm0,xmm0
      mulss    xmm0,.invdet
      movss    xmm1,xmm0
      movss    .u,xmm0
      cmpnless xmm1,[epsone]
      cmpnless xmm0,[epsminus]
      pxor     xmm1,xmm0
      movd     eax,xmm1
      or       eax,eax
      jz       @f

      lea      esi,.tvec
      lea      edi,.e1
      lea      ebx,.qvec
      call     cross_aligned

      movaps   xmm0,.dir
      mulps    xmm0,.qvec
      haddps   xmm0,xmm0
      haddps   xmm0,xmm0
      mulss    xmm0,.invdet
      movss    .v,xmm0
      movss    xmm1,xmm0
      addss    xmm1,.u
      cmpnless xmm1,[epsone]
      cmpnless xmm0,[epsminus]
      pxor     xmm1,xmm0
      movd     eax,xmm1
      or       eax,eax
      jz       @f

      movaps   xmm1,.e2
      mulps    xmm1,.qvec
      haddps   xmm1,xmm1
      haddps   xmm1,xmm1
      mulss    xmm1,.invdet
      movss    .t,xmm1
   ;   cmpnless xmm1,[eps]
   ;   movmskps eax,xmm1
   ;   test     eax,1
   ;   jz       @f
      comiss   xmm1,[eps]
      jl       @f

      mov      eax,1
      cmp      .ift,0
      je       .end       ; ok intersect occured, no edge cause

      movss    xmm0,.t      ; else check with edge lenght
    ;  movss    xmm1,.t
      cmpnless xmm0,[eps]
      cmpnless xmm1,.ed_l
      xorps    xmm0,xmm1
      movd     ebx,xmm0
      or       ebx,ebx
      jz       @f

 ;     mov      eax,1
 ;     movaps   xmm0,.t
      jmp      .end
   @@:
      xor      eax,eax
    .end:
      movaps   xmm0,.t
      add      esp,220
      pop      ebp
ret



;==================================================================
if 0
calc_intersection_point:
; p1, p2,p3 - tri vertices
; q1, q2 - bgn, end of ray(edge)
;   in:
;        xmm0 - q1
;        xmm1 - q2
;        xmm2 - p1
;        xmm5 - N  -    Normal vect of tri
;   out:
;        xmm0 - intersection point
     push    ebp
     mov     ebp,esp
     and     ebp,-16
     sub     esp,90
    .q1 equ  [ebp-16]
    .q2 equ  [ebp-32]
    .p1 equ  [ebp-48]
    .n  equ  [ebp-64]
    andps    xmm5,[zero_hgst_dd]
    movaps   .q1,xmm0
    movaps   .q2,xmm1
    movaps   .p1,xmm2
    movaps   .n,xmm5
    subps    xmm0,xmm2
    mulps    xmm0,xmm5
    haddps   xmm0,xmm0
    haddps   xmm0,xmm0
    subps    xmm1,.q1
    mulps    xmm1,xmm5
    haddps   xmm1,xmm1
    haddps   xmm1,xmm1
    divps    xmm0,xmm1
    movaps   xmm1,xmm0
    xorps    xmm1,[sign_mask]
    andps    xmm0,[abs_mask]
    andps    xmm1,[sign_mask]
    orps     xmm0,xmm1         ; xmm0 - t
    movaps   xmm2,.q2
    subps    xmm2,.q1
    mulps    xmm0,xmm2
    addps    xmm0,.q1         ; xmm0 - intersection point
    andps    xmm0,[zero_hgst_dd]
    add      esp,90
    pop      ebp
ret
end if
;====================================================================
mark_colided_edges:
; Proc has function of smothing inside edges also
;   if eax = 1 - smooth inside edges
   push    eax

   cmp     eax,1
   je      @f
   prompt  prompt_det_coll_ed  ; macro
   jmp     .f
  @@:
   prompt  prompt_smooth_in_ed
  .f:
   mov     [thread_params+4],eax
   mov     ebx,[points_count_var]
   add     ebx,80
   imul    ebx,[i12]
   invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   mov     [thread_params],eax
   mov     edi,eax
   mov     ecx,[points_count_var]
   lea     ecx,[ecx*3]
   mov     eax,-1
   cld
   rep     stosd

;   call  th_coll_ed0
;   call  th_coll_ed1
;   call  th_coll_ed2
;   call  th_coll_ed3

if 1
     invoke CreateThread,NULL,NULL,th_coll_ed0,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID0]
     mov    [hthread0],eax
     invoke SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL

     invoke CreateThread,NULL,NULL,th_coll_ed1,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID1]
     mov    [hthread1],eax
     invoke SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL

     invoke CreateThread,NULL,NULL,th_coll_ed2,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID2]
     mov    [hthread2],eax
     invoke SetThreadPriority,hthread2,THREAD_PRIORITY_TIME_CRITICAL

     invoke CreateThread,NULL,NULL,th_coll_ed3,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID3]
     mov    [hthread3],eax
     invoke SetThreadPriority,hthread3,THREAD_PRIORITY_TIME_CRITICAL

     invoke SetThreadAffinityMask,[hthread0],0 ;1
     invoke SetThreadAffinityMask,[hthread1],0 ;2
     invoke SetThreadAffinityMask,[hthread2],0 ;3
     invoke SetThreadAffinityMask,[hthread3],0 ;4

     invoke ResumeThread,[hthread0]
     invoke ResumeThread,[hthread1]
     invoke ResumeThread,[hthread2]
     invoke ResumeThread,[hthread3]

     invoke WaitForMultipleObjectsEx,4,hthread0,1,-1,0

     invoke CloseHandle,[hthread0]
     invoke CloseHandle,[hthread1]
     invoke CloseHandle,[hthread2]
     invoke CloseHandle,[hthread3]
end if

     mov     esi,[thread_params]
     mov     edi,[points_r_ptr]
     cld
     mov     ecx,[points_count_var]
   .update:
     cmp     [esi],dword -1
     je      @f
     movlps  xmm0,[esi]
     movss   xmm1,[esi+8]
     movlps  [edi],xmm0
     movss   [edi+8],xmm1
    @@:
     add     esi,12
     add     edi,12
     loop    .update

     pop    eax
     cmp    eax,1
     je     .skip_cnt  ; smooth ins edges cause

    mov    ecx,[edges_count]
    shr    ecx,3
    inc    ecx
    mov    esi,[edges_intersect_ptr]
    xor    ebx,ebx
   .pop_count:
    push   ecx
    xor    ecx,ecx
   @@:
    mov    eax,[esi]
    shr    eax,cl
    and    eax,1b
    add    ebx,eax
    inc    ecx
    cmp    ecx,8
    jne    @b
    inc    esi
    pop    ecx
    loop   .pop_count
    mov    [edges_coll_count],ebx

   .skip_cnt:
    invoke  VirtualFree ,dword[thread_params],0 , MEM_RELEASE

      cls   ; macro
ret

th_coll_ed0:
   mov   eax,0
   call  mark_colided_edges_th
ret
th_coll_ed1:
   mov   eax,1
   call  mark_colided_edges_th
ret
th_coll_ed2:
   mov   eax,2
   call  mark_colided_edges_th
ret
th_coll_ed3:
   mov   eax,3
   call  mark_colided_edges_th
ret


mark_colided_edges_th:
;in: - some globals
;      dword[thread_params] - ptr to modified vertices list
;      dword[thread_params+4] = 1 -> smooth ins ed, otherwise mark
;      collided edges
;      eax - thread No
;      [edges_intersect_ptr] - colided tri list, each tri as bit,
;      each colided edge marked as 1, otherwise -> 0
   push    ebp
   mov     ebp,esp
   sub     esp,300
   and     ebp,-16

   .ed_coll_ptr equ dword[ebp-4]
   .tri_cnt      equ dword[ebp-8]
   .msk1         equ dword[ebp-12]
   .collEdCnt   equ dword[ebp-16]
   .p1t          equ [ebp-32]    ; triangles coords (vertices)
   .p2t          equ [ebp-48]
   .p3t          equ [ebp-64]
   .vertEd1      equ [ebp-80]
   .vertEd2      equ [ebp-96]
   .indexEd1     equ [ebp-112]
   .indexEd2     equ [ebp-128]
   .vertEd1new   equ [ebp-128-16]
   .th_No        equ dword[ebp-152]
   .edStart      equ [ebp-156]
   .edEnd        equ [ebp-160]
   .ed_norm      equ [ebp-176]
   .tri_norm     equ [ebp-192]
   .smooth_ed    equ [ebp-196]
   .EdLen        equ [ebp-200]
;   .index_vert   equ [ebp-200]  ; ind of new vert (from triang edges)
;   .index_tri    equ [ebp-204]  ; index of new tris (from triangulated ed)
   .value        equ [ebp-240]
   .normEd       equ [ebp-256]

   mov     ebx,0.008889
   movd    xmm0,ebx
   shufps  xmm0,xmm0,0
   movaps  .value,xmm0    ; margin value

   mov     .smooth_ed,ebx
   mov     .th_No,eax
   mov     ebx,[edges_count]
   mov     ecx,[edges_count]
   shr     ebx,2
   and     ecx,11b
   imul    eax,ebx
   mov     .edStart,eax
   add     eax,ebx
   cmp     .th_No,3
   jnz     @f
   add     eax,ecx
 @@:
   mov     .edEnd,eax

   mov     esi,[edges_ptr]
   mov     ebx,.edStart
   shl     ebx,3
   add     esi,ebx
   mov     ecx,.edStart
 .nx_ed:   ; next edge
   push    esi
   push    ecx
   mov     eax,[esi]
   mov     ebx,[esi+4]
   movd    xmm0,[esi]
   movd    xmm1,[esi+4]
   pshufd  xmm0,xmm0,0
   pshufd  xmm1,xmm1,0
   movdqa  .indexEd1,xmm0
   movdqa  .indexEd2,xmm1
   imul    eax,[i12]
   imul    ebx,[i12]
   add     eax,[points_r_ptr]
   add     ebx,[points_r_ptr]
   movups  xmm1,[ebx]
   movups  xmm2,[eax]
   andps   xmm2,[zero_hgst_dd]
   andps   xmm1,[zero_hgst_dd]
   movaps  .vertEd1,xmm2
   movaps  .vertEd2,xmm1
   subps   xmm2,xmm1
   movaps  xmm3,xmm2
   mulps   xmm2,xmm2
   haddps  xmm2,xmm2
   haddps  xmm2,xmm2
   sqrtps  xmm2,xmm2
   movss   .EdLen,xmm2
   rcpps   xmm2,xmm2
   mulps   xmm3,xmm2
   andps   xmm3,[zero_hgst_dd]
   movaps  .normEd,xmm3
   mov     edi,[triangles_ptr]
   mov     esi,[triangles_normals_ptr]
   mov     .tri_cnt,0
 .nx_tri:  ; next triangle
   push    edi
   push    esi
   movdqu  xmm0,[edi]
   movdqu  xmm1,[edi]
   pcmpeqd xmm0,.indexEd1
   pcmpeqd xmm1,.indexEd2
   por     xmm1,xmm0
   movmskps eax,xmm1
   and     eax,111b
 ;  popcnt  eax,eax
 ;  cmp     eax,2
   or      eax,eax
   jnz     .end_l_tri

   ; Check  dot pr of normal vect of tri and
   ; edge (ray), if its near zero - skip - no intersection.

;   movups  xmm1,[esi]
;   mulps   xmm1,.normEd
;   haddps  xmm1,xmm1
;   haddps  xmm1,xmm1
;   andps   xmm1,[abs_mask]
;   comiss  xmm1,[eps]
;   jb     .end_l_tri

   mov     eax,[edi]
   mov     ebx,[edi+4]
   mov     edx,[edi+8]
   imul    eax,[i12]
   imul    ebx,[i12]
   imul    edx,[i12]
   add     eax,[points_r_ptr]
   add     ebx,[points_r_ptr]
   add     edx,[points_r_ptr]

   movups   xmm2,[eax]
   movups   xmm3,[ebx]
   movups   xmm4,[edx]
   movaps   .p1t,xmm2     ; .p1t - tri 1st vertex
   movaps   .p2t,xmm3
   movaps   .p3t,xmm4
;   movaps   xmm0,.vertEd1
;   movaps   xmm4,.vertEd1
;   movaps   xmm5,xmm1
;   minps    xmm0,.vertEd2
;   maxps    xmm4,.vertEd2
;   maxps    xmm5,xmm2
;   maxps    xmm5,xmm3
;   minps    xmm1,xmm2
;   minps    xmm1,xmm3       ; Tmax>Emin and Tmin<Emax
;   cmpnltps xmm5,xmm0
;   cmpltps  xmm1,xmm4
;   andps    xmm5,xmm1
;   movmskps eax,xmm5
;   and      eax,111b
;   cmp       eax,111b
;   jne      .end_l_tri


;intersect_tri: procs header
; in:
;     xmm0 - ray direction  ; should be normalized
;     xmm1 - ray orgin
;     xmm2 - tri vert1
;     xmm3 - tri vert2
;     xmm4 - tri vert3
;     if  eax = 1 - intersction with edge
;        xmm6 - edge lenght
;     if  eax = 0 - intersect with ray (classic)
; out:
;     eax  = 1 - intersection occured
;     xmm0 - float lo -> hi = t, v, u, ...
 @@:
   movaps    xmm0,.normEd
   movaps    xmm1,.vertEd2
   movaps    xmm2,.p1t
   movaps    xmm3,.p2t
   movaps    xmm4,.p3t
   movss     xmm6,.EdLen
   mov       eax,1
   call      intersect_tri

 .skip:
   cmp     eax,1
   je     .inter

  .end_l_tri:
   pop     esi
   pop     edi
   add     esi,12
   add     edi,12
   inc     .tri_cnt
   mov     edx,.tri_cnt
   cmp     edx,[triangles_count_var]
   jnz     .nx_tri
  .nx_loop_end:
   pop     ecx
   pop     esi
   add     esi,8
   inc     ecx
   cmp     ecx,.edEnd
   jne     .nx_ed
   jmp     .end


 .inter:
 ;  [edi] - tri
 ;  [esi] - edge
   cmp    dword[thread_params+4],1  ; no mark
   je     @f
   mov    eax,ecx  ;.tri_cnt
   mov    ebx,ecx  ;.tri_cnt
   shr    eax,3
   and    ebx,111b
   add    eax,[edges_intersect_ptr]
   mov    esi,[eax]
   bts    esi,ebx
   lock   xchg [eax],esi
  @@:

   cmp    dword[thread_params+4],1
   jne    .nx_end   ; check smooth inside edges cause

   mov    ebx,.tri_cnt
   imul   ebx,[i12]
   add    ebx,[triangles_normals_ptr]
   movups  xmm5,[ebx]
   movaps  .tri_norm,xmm5
   shufps  xmm0,xmm0,0  ; xmm0 - t
   mulps   xmm0,.normEd
   addps   xmm0,.vertEd2
  ; xmm0 -- intersection point coords
   movaps  xmm7,xmm0
   subps   xmm0,.vertEd1
   mulps   xmm0,.tri_norm
   haddps  xmm0,xmm0
   haddps  xmm0,xmm0 ; dot
   movmskps eax,xmm0

   movaps  xmm0,.normEd
   mulps   xmm0,.value ; remains margin value

   cmp     eax,0xf
   je      @f
   addps   xmm7,xmm0
   mov     ebx,.indexEd1
   jmp     .fix
  @@:
   mov     ebx,.indexEd2
   subps   xmm7,xmm0
  .fix:
   imul    ebx,[i12]
   movhlps xmm6,xmm7
   add     ebx,[thread_params]
   movlps  [ebx],xmm7
   movss   [ebx+8],xmm6

  .nx_end:
   pop    esi
   pop    edi
   jmp    .nx_loop_end
 .end:
;   mov     ebx,.ed_coll_ptr
;   mov     ecx,.collEdCnt       ; Colided edges count



   add     esp,300
   pop     ebp
ret

;===============================================================
if 0
find_intersection:
; in:
;   xmm0 - ray  orgin
;   xmm4 - ray  end
;   xmm1 - triangle first vertex_
;   xmm2 - tri second vert
;   xmm3 - tri 3th vert
; out:
;   eax = 1 intersection occured
;   eax = 0 no intersection

   push    ebp
   mov     ebp,esp
   sub     esp,104
   and     ebp,-16
   .p1t       equ [ebp-16]
   .p2t       equ [ebp-16*2]
   .p3t       equ [ebp-16*3]
   .vert_Ed1  equ [ebp-16*4]
   .vert_Ed2  equ [ebp-16*5]
   .msk1      equ [ebp-16*5-4]


   movaps  .vertEd1,xmm0
   movaps  .vertEd2,xmm4
   movaps  .p1t,xmm1     ; .p1t - tri 1st vertex
   movaps  .p2t,xmm2
   movaps  .p3t,xmm3
   movaps  xmm0,.vertEd1
   call    signed_vol_tetra
   movmskps eax,xmm4
   mov     .msk1,eax

   movaps  xmm1,.p1t
   movaps  xmm2,.p2t
   movaps  xmm3,.p3t
   movaps  xmm0,.vertEd2
   call    signed_vol_tetra
   movmskps eax,xmm4
   xor     eax,.msk1
   and     eax,1b
   cmp     eax,0
   je      .no_inter

   movaps  xmm0,.vertEd1
   movaps  xmm1,.vertEd2
   movaps  xmm2,.p1t
   movaps  xmm3,.p2t
   call    signed_vol_tetra
   movmskps eax,xmm4
   mov     .msk1,eax

   movaps  xmm0,.vertEd1
   movaps  xmm1,.vertEd2
   movaps  xmm2,.p2t
   movaps  xmm3,.p3t
   call    signed_vol_tetra
   movmskps eax,xmm4
   xor     eax,.msk1
   and     eax,1b
   cmp     eax,1
   je      .no_inter

   movaps  xmm0,.vertEd1
   movaps  xmm1,.vertEd2
   movaps  xmm2,.p3t
   movaps  xmm3,.p1t
   call    signed_vol_tetra
   movmskps eax,xmm4
   xor     eax,.msk1
   and     eax,1b
   cmp     eax,0
   jne     .no_inter

   mov     eax,1
   jmp     .end
  .no_inter:
   xor      eax,eax
  .end:
   add     esp,104
   pop     ebp
ret
;=======================================================
signed_vol_tetra:
 ; dot(cross(b-a,c-a),d-a)
 ; in:
 ;  xmm0 - a
 ;  xmm1 - b
 ;  xmm2 - c
 ;  xmm3 - d
 ; out:
 ;  xmm4 - broadcasted dot

 .vec1 equ [ebp-16]
 .vec2 equ [ebp-32]
 .vec3 equ [ebp-48]
 .vec4 equ [ebp-64]

  push    ebp
  mov     ebp,esp
  sub     esp,100
  and     ebp,-16
  subps   xmm1,xmm0
  subps   xmm2,xmm0
  subps   xmm3,xmm0
  movaps  .vec1,xmm1
  movaps  .vec2,xmm2
  movaps  .vec3,xmm3
  lea     esi,.vec1
  lea     edi,.vec2
  lea     ebx,.vec4
  call    cross_aligned
  movaps  xmm4,.vec4
  mulps   xmm4,.vec3
  andps   xmm4,[zero_hgst_dd]
  haddps  xmm4,xmm4
  haddps  xmm4,xmm4

  add     esp,100
  pop     ebp
ret
end if
;=============================================================

remove_dead_tri:
;  push  ebp
;  mov   ebp,esp
  mov     eax,-1
  movd    xmm7,eax
  pshufd  xmm7,xmm7,0

  mov     esi,[triangles_ptr]
  mov     ecx,[triangles_count_var]
 .chck:
  mov     eax,[esi]
  mov     ebx,[esi+4]
  mov     edx,[esi+8]
  cmp     eax,[esi+4]
  je      .tri_fail
  cmp     eax,[esi+8]
  je      .tri_fail
  cmp     ebx,[esi]
  je      .tri_fail
  cmp     ebx,[esi+8]
  je      .tri_fail
  cmp     edx,[esi]
  je      .tri_fail
  cmp     edx,[esi+4]
  je      .tri_fail
  add     esi,12
  loop    .chck
  jmp     .cop
 .tri_fail:
  movq    [esi],xmm7
  movd    [esi+8],xmm7
  add     esi,12
  loop    .chck
 .cop:
  mov     esi,[triangles_ptr]
  mov     edi,[triangles_ptr]
  mov     ecx,[triangles_count_var]
  xor     edx,edx
 .cp:
  mov     eax,[esi]
  cmp     eax,-1
  je      @f
  movdqu  xmm0,[esi]
  movq    [edi],xmm0
  movhlps xmm0,xmm0
  movd    [edi+8],xmm0
  add     edi,12
  inc     edx
 @@:
  add     esi,12
  loop    .cp
  mov     [triangles_count_var],edx
ret
;========================================================
remove_cracks:
   push  ebp
   mov   ebp,esp
; proc look through edge list, search  for edge E
; that in half lenght have some vertex, if found such
; vertex V search for triangle, that contains edge E and
; divide such triangle into two triangles that contains
; V, other words remove edge E.
  sub     esp,130
  and     ebp,-16
  .aaprox      equ [ebp-16]
  .Vi         equ [ebp-20]
  .Ei         equ dword[ebp-24]
  .tri11_ptr  equ dword[ebp-28]
  .tri12_ptr  equ dword[ebp-32]
  .crack_cnt  equ dword[ebp-36] ; crack counter
  .vec1       equ [ebp-64]
  .vec2       equ [ebp-80]
  .normal     equ [ebp-96]
  .tri13      equ [ebp-104]


  prompt  prompt_rm_cracks

  ; calc max count
  mov     eax,[edges_count]
  imul    eax,[i12]
  invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
  mov     .tri11_ptr,eax

  mov     eax,0.00005   ;0.000003
  movd    xmm0,eax
  pshufd  xmm0,xmm0,11000000b
  movdqa  .aaprox,xmm0
;  mov     eax,0.07
;  movd    xmm0,eax
;  pshufd  xmm0,xmm0,11000000b
;  movdqa  .aaprox2,xmm0
  mov     .crack_cnt,0

  mov     esi,[edges_ptr]
  xor     ecx,ecx
  mov     .Ei,ecx
 .srch:
  push    esi
  mov     eax,[esi]
  mov     ebx,[esi+4]
  imul    eax,[i12]
  imul    ebx,[i12]
  add     eax,[points_r_ptr]
  add     ebx,[points_r_ptr]
  movups  xmm0,[eax]
  movups  xmm1,[ebx]
  addps   xmm0,xmm1
  mulps   xmm0,[f05x3]

  xor     ecx,ecx
  mov     edi,[points_r_ptr]
  movaps  xmm2,xmm0
  movaps  xmm3,xmm0
  addps   xmm2,.aaprox
  subps   xmm3,.aaprox

 @@:
  movups  xmm1,[edi]
  movups  xmm4,[edi]
  cmpltps xmm1,xmm2
  cmpltps xmm4,xmm3
  xorps   xmm1,xmm4
  movmskps eax,xmm1
  and     eax,111b
  cmp     eax,111b
  je      .found
  add     edi,12
  inc     ecx
  cmp     ecx,[points_count_var]
  jng     @b
  pop     esi
  add     esi,8
  inc     .Ei
  mov     eax,.Ei
  cmp     eax,[edges_count]
  jne     .srch
  jmp     .end_clean


 .found:
;  eax   - E index in [edges_ptr]
;  ecx   - V index in [points_r_ptr]
  mov     .Vi,ecx
  mov     eax,.Ei
  shl     eax,3
  add     eax,[edges_ptr]
  movq    xmm0,[eax]
  movlhps xmm0,xmm0
if 0
  mov     esi,[triangles_ptr]
  mov     ecx,[triangles_count_var]
  xor     edi,edi
  movdqa  xmm1,xmm0 ; xmm0 - edge
  movdqa  xmm2,xmm0
  movd    xmm3,.Vi  ; ebx - middle vertex
  pshufd  xmm1,xmm1,0
  pshufd  xmm2,xmm2,01010101b
  pshufd  xmm3,xmm3,0
 @@:
  movups   xmm4,[esi]
  movups   xmm5,[esi]
  movaps   xmm6,xmm4
  pcmpeqd  xmm4,xmm1
  pcmpeqd  xmm5,xmm2
  pcmpeqd  xmm6,xmm3
  por      xmm5,xmm4
  por      xmm6,xmm5
  pmovmskb eax,xmm6
  and      eax,0xfff
  inc      edi
  add      esi,12
  cmp      eax,0xfff
  je       @f
  loop     @b
  jmp      .TiVL_done
 @@:
 int3
  pxor     xmm3,xmm3       ; mark triangle that contains
  movq     [esi],xmm3      ; this edge & vertex index
  movd     [esi+8],xmm3
;  mov      .TiV,edi
end if

 .TiVL_done:

  mov     esi,[triangles_ptr]
  mov     ecx,[triangles_count_var]
  movdqa  xmm4,xmm0
  movdqa  xmm5,xmm0
  pshufd  xmm4,xmm4,0
  pshufd  xmm5,xmm5,01010101b
  xor     edi,edi
 .bb:
  movdqu  xmm1,[esi]
  movdqu  xmm2,[esi]
  movdqa  xmm3,xmm1
  pcmpeqd xmm1,xmm4
  pcmpeqd xmm2,xmm5
  por     xmm2,xmm1
  movmskps eax,xmm2
  and      eax,111b
  mov      ah,al
  mov      bl,al
  shr      ah,1
  shr      bl,2
  and      al,1b
  and      bl,1b
  and      ah,1b
  add      bl,al
  add      bl,ah
;  popcnt   ebx,eax    ; if sse4 present
  cmp      bl,2
  je       @f
  add     esi,12
  inc     edi
  loop    .bb
  jmp     .oop    ; sth wrong
 @@:
  push     edi  ; tri index

;  esi - index of triangle
;  divide this tri into two triangles

  movups  xmm4,[esi]
  movd    xmm5,.Vi
  movd    [esi+8],xmm0     ; xmm0 - edge
  movd    [esi],xmm5       ; first triangle to triangles_ptr list
  movdqa  xmm7,xmm0
  movdqa  xmm6,xmm0
  pshufd  xmm7,xmm7,0
  pshufd  xmm6,xmm6,01010101b
  pcmpeqd xmm7,xmm4      ; xmm4 - old value of triangle
  pcmpeqd xmm6,xmm4
  por     xmm7,xmm6
  pxor    xmm7,[zero_hgst_dd]
  movdqa  xmm6,xmm7
  pand    xmm7,xmm4
  movd    eax,xmm6
  or      eax,eax
  jnz     .ok
  psrldq  xmm6,4
  psrldq  xmm7,4
  movd    eax,xmm6
  or      eax,eax
  jnz     .ok
  psrldq  xmm6,4
  psrldq  xmm7,4
 .ok:
  movd    [esi+4],xmm7

  mov     edi,.crack_cnt
  imul    edi,[i12]
  add     edi,.tri11_ptr   ; second triangle to .tri1_ptr list
  psrldq  xmm0,4
  movd    [edi],xmm5       ; xmm5 Vi   -> indexes in
  movd    [edi+4],xmm7     ; xmm7      -> both triangles
  movd    [edi+8],xmm0     ;
  ; check normal vector
  ; if normal vector of triangles not equal -
  ; - inverse vertices declaration order
  mov     .tri13,edi
  pop     edi              ; tri index
  push    esi
  imul    edi,[i12]
  add     edi,[triangles_normals_ptr]
  movups  xmm7,[edi]       ; source tri normal
                           ; 1 st tri normal
  mov     eax,[esi]
  mov     ebx,[esi+4]
  mov     edx,[esi+8]
  imul    eax,[i12]
  imul    ebx,[i12]
  imul    edx,[i12]
  add     eax,[points_r_ptr]
  add     ebx,[points_r_ptr]
  add     edx,[points_r_ptr]
  movups  xmm0,[eax]
  movups  xmm1,[ebx]
  movups  xmm2,[edx]

  subps   xmm1,xmm0
  subps   xmm0,xmm2
  movaps  .vec1,xmm1
  movaps  .vec2,xmm0
  lea     esi,.vec1
  lea     edi,.vec2
  lea     ebx,.normal
  call    cross_product
  lea     edi,.normal
  call    normalize_vector
  movaps  xmm0,xmm7
  mulps   xmm0,.normal
  andps   xmm0,[zero_hgst_dd]
  haddps  xmm0,xmm0
  haddps  xmm0,xmm0
  movmskps eax,xmm0
  pop     esi
  bt      eax,0
  jc      @f
  movq    xmm6,[esi]
  pshufd  xmm6,xmm6,11100001b
  movq    [esi],xmm6
 @@:

  mov     edi,.tri13
                     ; 2cond tri normal
  mov     eax,[edi]
  mov     ebx,[edi+4]
  mov     edx,[edi+8]
  imul    eax,[i12]
  imul    ebx,[i12]
  imul    edx,[i12]
  add     eax,[points_r_ptr]
  add     ebx,[points_r_ptr]
  add     edx,[points_r_ptr]
  movups  xmm0,[eax]
  movups  xmm1,[ebx]
  movups  xmm2,[edx]

  subps   xmm0,xmm1
  movaps  .vec1,xmm0
  subps   xmm2,xmm1
  movaps  .vec2,xmm2
  lea     esi,.vec1
  lea     edi,.vec2
  lea     ebx,.normal
  call    cross_product
  lea     edi,.normal
  call    normalize_vector
  movaps  xmm0,xmm7
  mulps   xmm0,.normal
  andps   xmm0,[zero_hgst_dd]
  haddps  xmm0,xmm0
  haddps  xmm0,xmm0
  movmskps eax,xmm0
  bt      eax,0
  jc      @f
  mov     edi,.tri13
  movq    xmm6,[edi]
  pshufd  xmm6,xmm6,11100001b
  movq    [edi],xmm6
 @@:

  inc     .crack_cnt
 .oop:
  pop     esi
  add     esi,8
  inc     .Ei
  mov     ecx,.Ei
  cmp     ecx,[edges_count]
  jnz     .srch
; clean up
 .end_clean:

  mov    eax,.crack_cnt
  or     eax,eax
  jz     .end
;  int3
  add    eax,[triangles_count_var]
  add    eax,100
  imul   eax,[i12]
  invoke VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
  mov    .tri12_ptr,eax

  mov    edi,eax
  mov    esi,[triangles_ptr]
  mov    ecx,[triangles_count_var]
  lea    ecx,[ecx*3]
  mov    edx,ecx
  shr    ecx,2
  and    edx,11b
 @@:
  movaps xmm0,[esi]
  movaps [edi],xmm0
  add    esi,16
  add    edi,16
  loop   @b
  cld
  mov    ecx,edx
  rep    movsd

  mov    esi,.tri11_ptr
  mov    ecx,.crack_cnt
 @@:
  movups xmm0,[esi]   ; new tris are (maybye) only a few
  movups [edi],xmm0
  add    esi,12
  add    edi,12
  loop   @b

  ; new list of triangles are ready
  ; now remove unused vertices and
  ; calc unique edges

  invoke  VirtualFree ,[triangles_ptr],0 , MEM_RELEASE
  push    .tri12_ptr
  pop     [triangles_ptr]
  mov     eax,.crack_cnt
  add     [triangles_count_var],eax

 .end:

  invoke  VirtualFree ,.tri11_ptr,0 , MEM_RELEASE
  cls
  add    esp,130
  pop    ebp
ret

;=========================================================

tri_fc_th0:
  xor    eax,eax
  movaps xmm0,xword[tri_area_x1]
  call   tri_fc_th_frame
  invoke SendMessage,hthread0 ,WM_FINISH,0,0
  invoke ExitThread,0
ret

tri_fc_th1:
  mov    eax,1
  movaps xmm0,xword[tri_area_x1]
  call   tri_fc_th_frame
  invoke SendMessage,hthread1 ,WM_FINISH,0,0
  invoke ExitThread,1
ret

tri_fc_th2:
  mov    eax,2
  movaps xmm0,xword[tri_area_x1]
  call   tri_fc_th_frame
  invoke SendMessage,hthread2 ,WM_FINISH,0,0
  invoke ExitThread,2
ret

tri_fc_th3:
  mov    eax,3               ;hwnd
  movaps xmm0,xword[tri_area_x1]
  call   tri_fc_th_frame
  invoke SendMessage,hthread3,WM_FINISH,0,0
  invoke ExitThread,3
ret
;================================================================
tri_fc_th_frame:
;   triangulize threaded proc
;   in: eax - th id
;       xmm0 - triang area as 4 dwords

   .tri_frame    equ  [ebp-16]
   .vert_rot3    equ  [ebp-32]
   .vert_rot2    equ  [ebp-44]
   .vert_rot1    equ  [ebp-56]
   .vert_counter equ  [ebp-60]
   .index3       equ  [ebp-64]
   .index2       equ  [ebp-68]
   .index1       equ  [ebp-72]
   push     ebp
   mov      ebp,esp
   sub      esp,150
   and      ebp,-16
   movaps   .tri_frame,xmm0

   emms
   mov      edi,[new_tri_ptr]  ; mem buffer
   mov      ecx,[triangles_count_var]
   mov      ebx,[triangles_count_var]
   and      ebx,0011b
   shr      ecx,2
   mov      edx,ecx
   imul     edx,eax
   imul     edx,[i12]
   cmp      eax,3
   jne      @f
   add      ecx,ebx
 @@:
   mov      esi,[triangles_ptr]
   add      esi,edx
   shl      edx,2
   add      edi,edx

   ; do new triangles list

  .look_for:
   push     ecx
   pushad
   mov      ecx,3
;   mov      .vert_counter,dword 0
  .search:    ; search through edges list indexes of
              ; new calculated vertices
   push     ecx
   cmp      ecx,3
   jne      @f
   mov      ebx,[esi]
   movq     mm0,[esi]
   cmp      ebx,[esi+4]
   jl       .do
   pshufw   mm0,mm0,01001110b
   jmp      .do
 @@:
   cmp      ecx,2
   jne      @f
   mov      ebx,[esi+4]
   movq     mm0,[esi+4]
   cmp      ebx,[esi+8]
   jl       .do
   pshufw   mm0,mm0,01001110b
   jmp      .do
 @@:
   movdqu   xmm0,[esi]
   mov      ebx,[esi]
   pshufd   xmm0,xmm0,00001000b
   movdq2q  mm0,xmm0
   cmp      ebx,[esi+8]
   jl      .do
   pshufw   mm0,mm0,01001110b

 .do:
   mov      ebx,[edges_ptr]
   xor      edx,edx      ; edx - edges counter == index
  .b:
   movq     mm2,mm0
   pcmpeqd  mm2,[ebx]
   pmovmskb eax,mm2
   cmp      eax,0xff
   je       .found
   inc      edx
   add      ebx,8
   cmp      edx,[edges_count]
   jnz      .b
  ; int3    ; sth not ok
  ; prompt -  object degenerated
  .found:                ; index of new vertex is: edge index
                         ; plus [points_count_var]
   add      edx,[points_count_var]
   mov      [edi],edx  ; index of vert temporaly to new triangles list
   ; rotate, scale and translate this vertex
   ; and check if it lay in triangulation frame
   add      edi,4
   pop      ecx
   loop     .search
   sub      edi,12

   xor      ecx,ecx
   push     esi
   push     edi
 @@:
   push     ecx
   push     edi
   shl      ecx,2
   mov      edx,[edi+ecx]
   shl      edx,2
   lea      esi,[edx*3]
   add      esi,[new_vert_ptr]
   lea      ebx,[matrix_scaled]
   lea      edi,.vert_rot1
   lea      ecx,[ecx*3]
   add      edi,ecx
   mov      ecx,1
   call     rotary
   pop      edi
   pop      ecx
   inc      ecx
   cmp      ecx,3
   jne      @b
   pop      edi
   pop      esi

   movups   xmm7,.vert_rot3
   movups   xmm6,.vert_rot2
   movups   xmm5,.vert_rot1
   movaps   xmm4,xmm7
   orps     xmm4,xmm6
   orps     xmm4,xmm5
   movmskps ebx,xmm4
   movaps   xmm0,.tri_frame
   movq     xmm1,[xxadd]
   cvtdq2ps xmm1,xmm1
   movlhps  xmm1,xmm1
   subps    xmm0,xmm1
   movhlps  xmm1,xmm0
   movaps   xmm4,xmm7
   cmpltps  xmm4,xmm0
   cmpltps  xmm7,xmm1
   xorps    xmm7,xmm4
   movmskps eax,xmm7
   and      eax,11b
   cmp      eax,11b
   jne      .no_in_f
   movaps   xmm4,xmm6
   cmpltps  xmm4,xmm0
   cmpltps  xmm6,xmm1
   xorps    xmm6,xmm4
   movmskps eax,xmm6
   and      eax,11b
   cmp      eax,11b
   jne      .no_in_f
   movaps   xmm4,xmm5
   cmpltps  xmm4,xmm0
   cmpltps  xmm5,xmm1
   xorps    xmm5,xmm4
   movmskps eax,xmm5
   and      eax,11b
   cmp      eax,11b
   jne      .no_in_f
;   jmp      .do_tri
;   ; vert of tri in frame area
   cmp      [td_wp_flag],0
   je       @f    ;whole, both pos & negative
                  ;  .vert_rot
   and      ebx,100b
   cmp      ebx,100b
   jne      .no_in_f
  @@:
   jmp      .do_tri


 .no_in_f:
;   cmp      dword .vert_counter,3
;   jmp      .do_tri   ; tri full in triangulize area
 @@:
   pxor     xmm5,xmm5
   movdqu   xmm4,[esi]
   movdqu   [edi],xmm4
   movdqu   [edi+12],xmm5
   movdqu   [edi+16+12],xmm5
   movd     [edi+32+12],xmm5
   jmp      .next
 .do_tri:
   movd     mm4,[edi]  ;   (x)mm4 - (x)mm6   ->  indexes
   movd     mm5,[edi+4]
   movd     mm6,[edi+8]
   movd     xmm4,[esi]
   movd     xmm5,[esi+4]
   movd     xmm6,[esi+8]

   movd     [edi],xmm4     ; indexes in new triangles list
   movd     [edi+4],mm4
   movd     [edi+8],mm6

   movd     [edi+12],mm4
   movd     [edi+16],mm5
   movd     [edi+20],mm6

   movd     [edi+24],mm4
   movd     [edi+28],xmm5
   movd     [edi+32],mm5

   movd     [edi+36],mm6
   movd     [edi+40],mm5
   movd     [edi+44],xmm6
  .next:
   popad
   add      edi,48
   pop      ecx
   add      esi,12
   sub      ecx,1
   jnz      .look_for


   add     esp,150
   pop     ebp

ret
;==============================================================
triangulize_all_faces:   ;all faces from teslate area
; in:
;  uses global variables that describes triangles, vertices and edges
;  eax - thread id
; out:
;  sets new global variables that describes triangles and vertices,
;  but edges parameters should be computed externally.
   push ebp
   mov  ebp,esp
   sub  esp,100
   and  ebp,0xfffffff0
   prompt  prompt_tri
   .new_vert_ptr equ [ebp-4]
   .new_tri_ptr  equ [ebp-8]
   .new_edge_ptr equ [ebp-12]
   .th_no        equ [ebp-16]
   .end_tri      equ [ebp-20]
   .tri_ptr      equ [ebp-24]
   emms
   mov    .th_no,eax
   mov    eax,[edges_count]
   add    eax,[points_count_var]
   add    eax,10
   shl    eax,2
   lea    eax,[eax*3]

   invoke VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   mov    .new_vert_ptr,eax   ; new vertices and old vertices
   mov    [new_vert_ptr],eax
   mov    esi,[points_r_ptr]
   mov    edi,eax
   mov    ecx,[points_count_var]
 @@:
   movups xmm0,[esi]
   movups [edi],xmm0
   add    esi,12
   add    edi,12
   loop   @b


   mov      esi,[edges_ptr]     ; calc all new vertices
   mov      ecx,[edges_count]
   mov      eax,0.5
   movd     xmm0,eax
   shufps   xmm0,xmm0,0
 .calc_mid_vert:
   mov      eax,[esi]
   mov      ebx,[esi+4]
   shl      eax,2
   shl      ebx,2
   lea      eax,[eax*3]
   lea      ebx,[ebx*3]
   add      eax,[points_r_ptr]
   add      ebx,[points_r_ptr]
   movups   xmm1,[eax]
   movups   xmm2,[ebx]
   subps    xmm1,xmm2
   mulps    xmm1,xmm0
   addps    xmm2,xmm1
   movups   [edi],xmm2
   add      esi,8
   add      edi,12
   sub      ecx,1
   jnz      .calc_mid_vert

   mov      eax,[triangles_count_var]
   add      eax,40
   shl      eax,4
   lea      eax,[eax*3]
   invoke VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   mov      .new_tri_ptr,eax
   mov      [new_tri_ptr],eax

;   mov      eax,[triangles_count_var]
;   add      eax,10
;   shl      eax,4
;   lea      eax,[eax*3]
;   invoke VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
;   mov      .tri_ptr,eax


    invoke CreateThread,NULL,NULL,tri_fc_th0,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID0]
    mov    [hthread0],eax
    invoke SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,tri_fc_th1,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID1]
    mov    [hthread1],eax
    invoke SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,tri_fc_th2,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID2]
    mov    [hthread2],eax
    invoke SetThreadPriority,hthread2,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,tri_fc_th3,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID3]
    mov    [hthread3],eax
    invoke SetThreadPriority,hthread3,THREAD_PRIORITY_TIME_CRITICAL

    invoke SetThreadAffinityMask,[hthread0],0 ;[thread_array]
    invoke SetThreadAffinityMask,[hthread1],0 ;[thread_array+4]
    invoke SetThreadAffinityMask,[hthread2],0 ;[thread_array+8]
    invoke SetThreadAffinityMask,[hthread3],0 ;[thread_array+12]

    invoke ResumeThread,[hthread0]
    invoke ResumeThread,[hthread1]
    invoke ResumeThread,[hthread2]
    invoke ResumeThread,[hthread3]

    invoke WaitForMultipleObjectsEx,4,hthread0,1,-1,0
    invoke CloseHandle,[hthread0]
    invoke CloseHandle,[hthread1]
    invoke CloseHandle,[hthread2]
    invoke CloseHandle,[hthread3]

   invoke  VirtualFree ,[points_r_ptr],0 , MEM_RELEASE
   invoke  VirtualFree ,[triangles_ptr],0 , MEM_RELEASE


   push      dword .new_vert_ptr
   pop       [points_r_ptr]


   mov       esi,.new_tri_ptr
   mov       edi,.new_tri_ptr
   mov       ecx,[triangles_count_var]
   shl       ecx,2
   xor       edx,edx
 .remove:
   cmp       dword[esi],0
   jne       @f
   cmp       dword[esi+4],0
   jne       @f
   cmp       dword[esi+8],0
   jne       @f
   add       esi,12
   loop      .remove
   jmp       .done
  @@:
   movdqu    xmm0,[esi]
   movq      [edi],xmm0
   psrldq    xmm0,8
   movd      [edi+8],xmm0
   add       edi,12
   add       edx,1
   add       esi,12
   loop      .remove
  .done:

   push   dword .new_tri_ptr
   pop    [triangles_ptr]

   mov    eax,[edges_count]
   add    [points_count_var],eax
;   mov    eax,[edges_count]
;   lea    eax,[eax*3]
;   mov    [edges_count],eax
;   mov    eax,[triangles_count_var]
;   shl    eax,2
   mov    [triangles_count_var],edx

   cls ;macro


   add  esp,100
   pop  ebp
ret
;====================================================================

do_variable_slices_ptr_buff:
; Procedure match every screen pixel with ptr, wich describes
; its depth history. If dword ptr = byte cnt = 0, not depth
; history, pixel has background color.
; in:  esi - counter buff
;      edi - ptr buff
;      eax - voxels depth history buff
;      ecx - count of all screen pixels
       jecxz  .end
     .again:
       movzx  edx,byte[esi]
       or     edx,edx
       jz     .zero
       mov    [edi],eax
       shl    edx,2
       add    eax,edx
       jmp    .u_cnts
     .zero:
       mov    [edi],dword 0
     .u_cnts:
       add    esi,1
       add    edi,4
       sub    ecx,1
       jnz    .again
    .end:
ret
;================================================
count_transparent_voxels:
; in:
;    edi - ptr to slices_cnt buffer
;          this buff describes how many
;          slices has every rendered pixel
;    eax - size x of screen
;    ebx - size y of scr
; out:
;    ecx - count of all transparent voxels
     push  ebp
   ;  int3
     mul   ebx
     xor   ebp,ebp
     shr   eax,2
     add   eax,1
     mov   esi,eax
   @@:
     movzx eax,byte[edi]
     movzx ebx,byte[edi+1]
     movzx ecx,byte[edi+2]
     movzx edx,byte[edi+3]
     add   ebp,eax
     add   ebp,ebx
     add   ebp,ecx
     add   ebp,edx
     add   edi,4
     sub   esi,1
     jnz   @b
     mov   ecx,ebp
     pop   ebp
ret
;======================================================
remove_unused_vertices:
; in:
;    eax - triangles (faces) ptr
;    ebx - points (vertices) ptr
;    ecx - triangles (faces) count
;    edx - points (vertices) count
; out:
;    changed points_ptr(ebx) list
;    ecx - new vertices count

    push  ebp
    mov   ebp,esp
    sub   esp,32

  .triangles_ptr           equ [ebp-4]
  .points_ptr              equ [ebp-8]
  .triangles_count         equ [ebp-12]
  .points_count            equ [ebp-16]
  .new_vert_with_old_i_ptr equ [ebp-20]
  .points_new_cnt          equ [ebp-28]

  mov   .triangles_ptr,eax
  mov   .points_ptr,ebx
  mov   .triangles_count,ecx
  mov   .points_count,edx

  prompt  prompt_unused_vert

  mov     eax,[points_count_var]
  shl     eax,4
  add     eax,65536
  invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
  or      eax,eax
  jnz     @f
  int3
 @@:
  mov    .new_vert_with_old_i_ptr,eax

;  movzx   eax,[triangles_count_var]
;  mul     [i6]
;  add     eax,65536
;  invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
;  or      eax,eax
;  jnz     @f
;  int3
; @@:
;   mov    .new_tri_list_ptr,eax

; od zm=0 do points_count rob - sprawdz czy w licie triangles
; jest zm jesli tak pisz dany vertex do listy .new_vert wraz z jego
; starym indeksem zm
;---------
; from zm=0 to points_count do - check if in triangles list
; is zm, if yes, write this vertex with old index zm to .new_vert list

  mov   dword .points_new_cnt,0
  mov   edi,.new_vert_with_old_i_ptr
  xor   eax,eax
  mov   edx,.triangles_count
  lea   edx,[edx*3]
 .find_index:
  mov   esi,.triangles_ptr
  xor   ecx,ecx
 .find:
  cmp   eax,[esi]
  je    .vert_exist
  add   esi,4
  add   ecx,1
  cmp   ecx,edx
  jnz   .find
  jmp   .loop_e ; vertex not exist in triangles list

 .vert_exist:
  mov    ebx,eax
  shl    ebx,2
  lea    ebx,[ebx*3]
  add    ebx,.points_ptr
  movups xmm0,[ebx]
  movlps  [edi],xmm0
  movhlps xmm0,xmm0
  movss   [edi+8],xmm0
  mov    [edi+12],eax
  add    edi,16
  add    dword .points_new_cnt,1
 .loop_e:
  add    eax,1
  cmp    eax,.points_count
  jnz    .find_index

; in the list pointed by .new_vert..._ptr are new
; vertices and its old index, list pointed by .triangles_ptr
; (.new_tri.._ptr) need renew too, instead old, it will be
; new indexes

  xor  eax,eax
  mov  edx,.triangles_count
  lea  edx,[edx*3]
  mov  edi,.new_vert_with_old_i_ptr
 .find_again:
  mov  ebx,[edi+12]
  mov  esi,.triangles_ptr
  xor  ecx,ecx
 .search:
  cmp  [esi],ebx
  jne  @f
  mov  [esi],eax
 @@:
  add  esi,4
  add  ecx,1
  cmp  ecx,edx
  jnz  .search
  add  edi,16
  add  eax,1
  cmp  eax,.points_new_cnt
  jnz  .find_again

  mov  esi,.new_vert_with_old_i_ptr    ;copy mem
  mov  edi,.points_ptr
  mov  ecx,.points_new_cnt

 @@:
  movups xmm0,[esi]
  movups [edi],xmm0
  add    esi,16
  add    edi,12
  sub    ecx,1
  jnz    @b

  invoke  VirtualFree ,.new_vert_with_old_i_ptr,0 , MEM_RELEASE
  cls
  mov   ecx,.points_new_cnt

  mov   esp,ebp
  pop   ebp
ret
